/****************************************************************************
*****************************************************************************
**   The Watcher extender                                                  **
**                                                                         **
**      Watch directories. Return when a file is created or modified.      **
**                                                                         **
*****************************************************************************
*****************************************************************************
**                                                                         **
**  By: Alan Kreutzer,  April 2003                                         **
**                                                                         **
**  Requires WinBatch version 2001 or later                                **
**                                                                         **
**  May be freely distributed.                                             **
**  This program is distributed with no warrenty, use at your own risk.    **
**                                                                         **
*****************************************************************************
****************************************************************************/
#include <windows.h>
#include "wwwpform.h"
#include "addons.h"


/*---------------------------------------------------------------------+
!  WATCHMAX defines how many directories we can watch simultaneously.  !
+---------------------------------------------------------------------*/
#define WATCHMAX 10

/*---------------------------------------------------------------------+
!  If the DEBUGALLOC flag is set to 1, then we keep a count of all     !
!  memory ALLOC and FREE calls, and display the total at the end.      !
!  (If the final total is not zero, we know there is a memory leak.)   !
+---------------------------------------------------------------------*/
#define DEBUGALLOC 0  

/*---------------------------------------------------------------------+
!  Defines for WinBatch.                                               !
+---------------------------------------------------------------------*/
#define INTERFACELEVEL  2           // extender interface level 
#define EXTENDER_FEATURE_LEVEL 1001 // extender feature level
#define STRICT

           


/*-----------------------------------------------------------------------+
!       Definitions used when manipulating arrays.                       !
+-----------------------------------------------------------------------*/
#define VARTYPE_UNDEF      0           // These are bitmapped, var can be both
#define VARTYPE_INT        1           // Numeric and string, must check bits
#define VARTYPE_STRING     2
#define VARTYPE_FILE       5           // FILE and INT bits set
#define VARTYPE_UNINIT     8
#define VARTYPE_OLEOBJECT 17           // OLE AND INT bits set
#define VARTYPE_FLOATNUM  32           // floating point
#define VARTYPE_BINARY    65           // BINARY and INT bits set
#define VARTYPE_XOLEOBJECT 129         /// an OLE object that has been temporarily camouflaged
#define VARTYPE_ARRAY     256          /// an array

typedef struct arrayvariable
{
  double Aflt;                         // 8-byte floating point number
  LONG   Ax;                           // if int variable, value is here
  int    Ainittype;                    // 0=uninitialized 1=int 2=string 3=both  yes can have both sometimes
  DWORD  Astrptr;                      // pointer to string space for string
} arrvar;


typedef struct resultants2
{
  union u
  {
    struct s                           // for normal variables
    {
      double flt;                      // 8-byte floating point number
      LONG   x;                        // if int variable, value is here
    };
    struct ss                          // for VARTYPE_ARRAY
    {
      WORD wDim[5];                    // array dimension sizes
      BYTE cDims;                      // number of array dimensions
      BYTE cFlags;                     // array flags
    };
  };
  int   inittype;                      // 0=uninitialized 1=int 2=string 3=both  yes can have both sometimes
  union uu
  {
    DWORD strptr;                      // pointer to string space for string
    LPSTR lpBuffer;                    // pointer to locked array data buffer for VARTYPE_ARRAY
  };
  WORD  vorigin;                       // index to var area is derived from var, or FFFF
} resultvar2;

typedef arrvar FAR *LPARRVAR;

typedef resultvar2 FAR * LPRESULTVAR2;


    /*------------------------------------------------------------------+
    !  This a generic "linked list".                                    !
    +------------------------------------------------------------------*/
typedef struct _LINKED
{
   struct _LINKED *Next;
   char            Buffer[4];
} LINKED;

    /*------------------------------------------------------------------+
    !   We use this block to build a linked list of all the files in    !
    !   a directory.   We build a chain for a directory at startup.     !
    !   When we get a signal that something has changed in the          !
    !   directory, we build another chain and then compare it to the    !
    !   original chain.  The differences tell us what changed.          !
    !                                                                   !
    !   We save the file modification time in the FileTime word.        !
    !   This can be used to tell if a file has been modified since      !
    !   the original chain was formed.   "Name" is the file name,       !
    !   including any subdirectory names.                               !   
    +------------------------------------------------------------------*/                                   
typedef struct _FILECHAIN 
{ 
    struct _FILECHAIN *Next;
    ULONG64     Size;
    ULONG64     Created;
    ULONG64     Modified;
    char        Name[4];
}FILECHAIN;

 
    /*------------------------------------------------------------------+
    !   The WATCH structure tells us what directories to watch.         !
    !   Note...                                                         !
    !    The WATCH structure is NOT a linked list.                      !
    !    We keep a table of Watch structures in the COMMON block.       !
    +------------------------------------------------------------------*/
typedef struct _WATCH              // Used to keep configuration information
{
    BOOL      Recursive;           // Watch subdirectories?
    BOOL      ChangeFlag;          // TRUE = Changes have been reported
    LINKED    *Only;               // Linked list of "Do Only" file names
    LINKED    *Ignore;             // Linked list of "Ignore" file names
    FILECHAIN *Files;              // chain of files under this directory
    long      DirLen;              // Length of Directory name
    char      Dir[2];              // The name of the directory we're watching.
} WATCH;

    /*------------------------------------------------------------------+
    !   This structure makes a linked list of information about files   !
    !   that have changed, and are to be reported to the user.          !
    !                                                                   !
    !   Size .... the file size.                                        !
    !   Created.. The creation date.  (in FILETIME format)              ! 
    !   Modified. Modification date.  (in FILETIME format)              !
    !   Watch.... The Watch Structure this file belongs to.             !
    !   NewFile.. 0 = Modified file,  1 = Newly created file.           !
    !   Name..... The file path (not including the base directory.)     !
    +------------------------------------------------------------------*/                                   
typedef struct _CHANGED 
{ 
    struct _CHANGED *Next;
    WATCH     *Watch;
    ULONG64   Size;
    ULONG64   Created;
    ULONG64   Modified;
    char      NewFileFlag;
    char      Name[3];
}CHANGED;

 
/*-----------------------------------------------------------------------+
!                                                                        !
!   Structure used to save data between calls.                           !
!   (We save the address of this structure in the lpLong variable,       !
!    so it will be available from one call to the next.)                 !
!                                                                        !
!   Note... the "Buffer" MUST be the first item in the structure,        !
!   because we sometimes use it to build arrays.  (WinBatch will not     !
!   accept an array unless its buffer begins at the very beginning of    !
!   a borrowed memory block.)                                            !
!                                                                        !
+-----------------------------------------------------------------------*/
typedef struct _COMMON
{
    char            Buffer[512];      /* MUST BE 1ST ITEM IN STRUCTURE        */
    char            Path[MAX_PATH*2]; /* Buffer to build qualified paths.     */
    WIN32_FIND_DATA w32find;          /* Workarea for FindNextFile()          */
    
    DLLSERVICES     DllServices;   /* Callback Routine for Array functions.   */
    HGLOBAL         hBatData;      /* Required parameter for Array Callback.  */
    
    double          Float;         /* Floating point(when Return = 'F'        */
    char           *String;        /* String        (when return = 'S')       */
    long            Long;          /* Numeric value (when return = 'L')       */
    long            ErrorLevel;    /* Error type... -1 (Normal) or -3 (Fatal) */
    resultvar2      Array;         /* Array         (when return = 'A')       */
    char            Return;        /* A=Array, S=String, L=Long, E=Error.     */
    
    CHANGED        *ChangeQueue;   /* Files to be reported to the user.       */
    
    long            WatchCount;        /* # of directories we're watching     */
    WATCH          *Watches[WATCHMAX]; /* Directories being watched           */
    HANDLE          Handles[WATCHMAX]; /* GetNextChangeNotifcation handles.   */
} COMMON;


    /*------------------------------------------------------------------+
    !  This structure is used as a workarea by the routines that        !
    !  build the file chains.                                           !
    +------------------------------------------------------------------*/ 
typedef struct _FILECHAINWORK
{
    COMMON          *Common;
    char            *Path;
    char            *PathAboveBase;
    long             LengthAboveBase;
    WIN32_FIND_DATA *FileInfo;
    BOOL             Recursive;
    LINKED          *Only;
    LINKED          *Ignore;
    FILECHAIN        Root;
} FILECHAINWORK;

/****************************************************************************
*                                                                           *
*                         Function Prototypes                               *
*                                                                           *
****************************************************************************/
BOOL WINAPI DllMain(HANDLE hInst, DWORD dwReason, LPVOID lpReserved);

void  WatchVersion(COMMON *Common, long Request);
void  WatchOpen   (COMMON *Common, char *Directory, BOOL Recursive, char *OnlyList, char *IgnoreList);
void  WatchWait   (COMMON *Common, long Timeout, BOOL DetailFlag);
void  WatchClose  (COMMON *Common);

LINKED *BuildOnlyIgnore(COMMON *Common, char *List);

void *Allocate         (COMMON *Common, long Size);
void  Free             (HGLOBAL Memory);
void  FreeWatch        (WATCH  *Watch);
void  FreeLinkedList   (LINKED *Linked);

long  CaseComp (char *First, char *Second);
BOOL  WildComp (char *String, char *Wild);

void  FAR PASCAL MBSN   (char *Title, long Number);
void  FAR PASCAL MBSX   (char *Title, long Number);
void  FAR PASCAL MBSS   (char *Title, char *Text);

FILECHAIN *CreateFileChain(COMMON *Common, WATCH *WatchMe);
void  GetFilesRecursive(FILECHAINWORK *Work, char *Append);
BOOL  FoundInLinkedList(char *Name, LINKED *List);
void  FileChainInsert(FILECHAINWORK *Work);
void  LookForChanges (COMMON *Common, WATCH *Watch);
void  SaveChanged    (COMMON *Common, WATCH *Watch, FILECHAIN *File, BOOL NewFileFlag);
void  ReturnChangesToUser(COMMON *Common, BOOL DetailFlag);
void  ReturnTimeoutToUser(COMMON *Common, BOOL DetailFlag);
BOOL  ArrayStringPUT (COMMON *Common, LPARRVAR lpArray, char *TheString); 

void  SetError(COMMON *Common, long ErrNo, long ErrLevel, char *Param1, char *Param2);
/*-------------------------------------------------------------------+
!             Case conversion routine                                !
+-------------------------------------------------------------------*/
#define MyToLower(p1) LOBYTE(LOWORD((DWORD) (CharLower((LPSTR) (DWORD) (BYTE) (p1)))))
#define MyToUpper(p1) LOBYTE(LOWORD((DWORD) (CharUpper((LPSTR) (DWORD) (BYTE) (p1)))))




/****************************************************************************
*                                                                           *
*                           Global Variables                                *
*                                                                           *
****************************************************************************/
      
/*-------------------------------------------------------------------+
! Superstitious data field to bypass/ignore bugs in some versions    !
! of Windows that cause the beginning of the data area to be         !
! overwritten.  Should be the first data area defined and linked.    ! 
+-------------------------------------------------------------------*/
char szBugFiller[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";

HINSTANCE   hinstDLL;                    // DLL Instance saved

#if DEBUGALLOC == 1
long        GlobalAllocCount = 0;
#endif

/*------------------------------------------------------+
!       Case conversion table.                          !
!       (Converts upper to lower case)                  !
+------------------------------------------------------*/
#define LOWERCASE(C) LowerTable[(unsigned char) C]
unsigned char LowerTable[256] =
    {0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0a,0x0b,0x0c,0x0d,0x0e,0x0f,
     0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17,0x18,0x19,0x1a,0x1b,0x1c,0x1d,0x1e,0x1f,
     0x20,0x21,0x22,0x23,0x24,0x25,0x26,0x27,0x28,0x29,0x2a,0x2b,0x2c,0x2d,0x2e,0x2f,
     0x30,0x31,0x32,0x33,0x34,0x35,0x36,0x37,0x38,0x39,0x3a,0x3b,0x3c,0x3d,0x3e,0x3f,

     0x40,0x61,0x62,0x63,0x64,0x65,0x66,0x67,0x68,0x69,0x6a,0x6b,0x6c,0x6d,0x6e,0x6f,
     0x70,0x71,0x72,0x73,0x74,0x75,0x76,0x77,0x78,0x79,0x7a,0x5b,0x5c,0x5d,0x5e,0x5f,
     0x60,0x61,0x62,0x63,0x64,0x65,0x66,0x67,0x68,0x69,0x6a,0x6b,0x6c,0x6d,0x6e,0x6f,
     0x70,0x71,0x72,0x73,0x74,0x75,0x76,0x77,0x78,0x79,0x7a,0x7b,0x7c,0x7d,0x7e,0x7f,
     0x80,0x81,0x82,0x83,0x84,0x85,0x86,0x87,0x88,0x89,0x9a,0x8b,0x9c,0x8d,0x8e,0x8f, 
     0x90,0x91,0x92,0x93,0x94,0x95,0x96,0x97,0x98,0x99,0x9a,0x9b,0x9c,0x9d,0x9e,0xFf, 
     0xA0,0xA1,0xA2,0xA3,0xA4,0xA5,0xA6,0xA7,0xA8,0xA9,0xAa,0xAb,0xAc,0xAd,0xAe,0xAf,
     0xB0,0xB1,0xB2,0xB3,0xB4,0xB5,0xB6,0xB7,0xB8,0xB9,0xBa,0xBb,0xBc,0xBd,0xBe,0xBf, 
     0xE0,0xE1,0xE2,0xE3,0xE4,0xE5,0xE6,0xE7,0xE8,0xE9,0xEa,0xEb,0xEc,0xEd,0xEe,0xEf,
     0xF0,0xF1,0xF2,0xF3,0xF4,0xF5,0xF6,0xD7,0xF8,0xF9,0xFa,0xFb,0xFc,0xFd,0xFe,0xDf,
     0xE0,0xE1,0xE2,0xE3,0xE4,0xE5,0xE6,0xE7,0xE8,0xE9,0xEa,0xEb,0xEc,0xEd,0xEe,0xEf,
     0xF0,0xF1,0xF2,0xF3,0xF4,0xF5,0xF6,0xF7,0xF8,0xF9,0xFa,0xFb,0xFc,0xFd,0xFe,0xFf}; 


/****************************************************************************
*                                                                           *
*                        Command Extension Table                            *
*                                                                           *
****************************************************************************/
tablestruct CommandExtTable[] =
{
//----------------------------------------------------------------------
// ID   Type       Mask              Params/Value, Size, Name
//----------------------------------------------------------------------
  { 1, CONSTANT, 0x0000000000000000,           0,  0, "@WatchPath"        },
  { 2, CONSTANT, 0x0000000000000000,           1,  0, "@WatchNew"         },
  { 3, CONSTANT, 0x0000000000000000,           2,  0, "@WatchSize"        },
  { 4, CONSTANT, 0x0000000000000000,           3,  0, "@WatchCreate"      },
  { 5, CONSTANT, 0x0000000000000000,           4,  0, "@WatchModify"      },
  {10, FUNCTION, 0x0000000000000000, 0x00000100L,  0, "WatchVersion"      }, 
  {11, FUNCTION, 0x0000000000001101, 0x00000004L,  0, "WatchOpen"         },
  {12, FUNCTION, 0x0000000000000000, 0x00000101L,  0, "WatchWait"         },
  {13, FUNCTION, 0x0000000000000000, 0x00000000L,  0, "WatchClose"        },
//--------------------------------------------------------------------------
  {-1,        0,                  0, 0x00000000,   0, ""                  }
};


/*******************************************************************************
********************************************************************************
**                                                                            **
**                                                                            **
**                                                                            **
**                                                                            **
**                               DllMain                                      **
**                                                                            **
**                                                                            **
**                                                                            **
**                                                                            **
********************************************************************************
*******************************************************************************/
BOOL WINAPI DllMain(HANDLE hInst, DWORD dwReason, LPVOID lpReserved)
{

  hinstDLL = hInst;
  UNREFERENCED_PARAMETER(dwReason);
  UNREFERENCED_PARAMETER(lpReserved);
#if DEBUGALLOC == 1
  if (dwReason == DLL_PROCESS_ATTACH) 
  {
      GlobalAllocCount = 0;
  }
  
  if (dwReason == DLL_PROCESS_DETACH)
  {
      char Buffer[128];
      wsprintf(Buffer,"At SHUTDOWN.  Allocated blocks = %d",GlobalAllocCount);
      MessageBox(GetFocus(),Buffer,"DEBUG",MB_OK);
  }
  
#endif
  return (TRUE);
}


/*******************************************************************************
********************************************************************************
**                                                                            **
**                                                                            **
**                                                                            **
**                                                                            **
**                        WILExtenderQuery                                    **
**                                                                            **
**           Tells the WilDll about our interface level.                      **
**                                                                            **
**                                                                            **
**                                                                            **
**                                                                            **
********************************************************************************
*******************************************************************************/
DllExport LONG FAR PASCAL WILExtenderQuery
 (short  nRequest,                     // type of information being requested
  LPVOID lpvInfo,                      // information buffer
  LONG   nCount)                       // additional data
{
  char szWork[256];
  LONG        i;

  switch (nRequest)
  {
/*----------------------------------------------------------------------+
!   case 1 ... identify yourself to a caller.                           !
+----------------------------------------------------------------------*/
    case 1:                      
    {
      long ExtenderVersion;
      
      if ((lpvInfo) && (nCount > 0))
      {
        LoadString(hinstDLL, 1000, szWork, sizeof(szWork));
        lstrcpyn((LPSTR) lpvInfo, szWork, nCount);
      }
      
      LoadString(hinstDLL, 1001, szWork, sizeof(szWork));   
      i = 0;
      ExtenderVersion = 0;
      while (szWork[i] != '\0')
      {
         ExtenderVersion = (ExtenderVersion * 10) + (szWork[i] - '0');
         i++;
      } 
      return ExtenderVersion;
    }

/*----------------------------------------------------------------------+
!   case 2 ... Tell WinBatch we use interface level 2                   !   
+----------------------------------------------------------------------*/
    case 2:                            // preferred interface version
    {
      for (i = 0; i < nCount; i++)
        if (((LPLONG) lpvInfo)[i] == INTERFACELEVEL)  // found one we like
          return (INTERFACELEVEL);

      return (-1);                     // didn't find any we like
    }


/*----------------------------------------------------------------------+
!    unknown requests                                           !  
+----------------------------------------------------------------------*/
    default:
      return (0);
  }
}




/*******************************************************************************
********************************************************************************
**                                                                            **
**                                                                            **
**                                                                            **
**                                                                            **
**                                                                            **
**                          WILExtender2                                      **
**                                                                            **
**                                                                            **
**                                                                            **
**           The main routine performs all extender functions,                **
**           and various initialization and cleanup tasks.                    **
**                                                                            **
**                                                                            **
**                                                                            **
**                                                                            **
**                                                                            **
********************************************************************************
*******************************************************************************/

DllExport LONG FAR PASCAL WILExtender2
  (HWND           hWnd,                   // hWnd of parent application
  HINSTANCE       hInst,                  // hInstance of parent application
  LONG            ViperInterfaceVersion,  // WIL interface version
  short           CmdId,                  // identifies extender function being called
  LPVIPERCALLBACK lpViperCallback,        // pointer to return value callback function
  LPVIPERSTRUCT   lpViperStruct,          // pointer to information structure
  LPLONG          lpLong,                 // pointer to your 32 bits
  LPVIPERVAR      lpvvArg,                // pointer to parameter array
  LONG            nArgCount,              // parameter count for lpvvArg
  LPVIPERVAR      lpViperReturn)          // pointer to return value structure
{  
    long    Ret;
    COMMON  *Common = (COMMON *) *lpLong;
    
     /***************************************************
     *                                                  *
     *    The "little Switch."                          *
     *                                                  *
     *    Calls with cmdId = -1, -2, or -3              *
     *    are for initialization and shutdown.          *
     *                                                  *
     ***************************************************/
  switch (CmdId)
  {

      /*-----------------------------------------------------------------+
      !  case -1 ... One time initialization routine.                    !
      !                                                                  !
      !         Allocate a block of memory for values that must be       !
      !         preserved between calls.                                 !
      !                                                                  !
      !         Return the number of functions  & constants defined.     !
      +-----------------------------------------------------------------*/
      case -1:
      {
         long X;
         
         /*------------------------------------------------------+
         !   Allocate the common block.                          !
         +------------------------------------------------------*/
         Common = (COMMON *) GlobalAlloc(GMEM_FIXED,sizeof(COMMON));
         if (Common == NULL)
         {
           LoadString(hinstDLL, 913, (char *) lpvvArg, nArgCount);
           return -1;
         }
         
#if DEBUGALLOC == 1
         GlobalAllocCount++;  // If we're debugging, count blocks allocated
#endif

         /*------------------------------------------------------+
         !   Finish initializing the common block.               !
         +------------------------------------------------------*/
         Common->WatchCount  = 0;
         Common->ChangeQueue = NULL;
         for (X=0; X<WATCHMAX; X++)
         {
             Common->Watches[X] = NULL;
             Common->Handles[X] = INVALID_HANDLE_VALUE;
         }
         *lpLong = (long) Common;  // Winbatch will save address for us.

         return ((sizeof(CommandExtTable) / sizeof(tablestruct)) - 1);
      }

      /*----------------------------------------------------------------+
      !  case -2 ... Called repeatedly as extender is being loaded.     !
      !                                                                 !
      !        On each call, return a function table entry.             !
      +----------------------------------------------------------------*/
      case -2:
      {
         char *Cx1;
         char *Cx2;
         long  CLen;
         LPTABLESTRUCT MyTable;
         LPTABLESTRUCT OutTable;

         MyTable = &CommandExtTable[nArgCount];

         if (MyTable->ident == -1) return (0);

         OutTable = (LPTABLESTRUCT) lpvvArg;
         if (OutTable)
         {
            OutTable->ident  = MyTable->ident;
            OutTable->type   = MyTable->type;
            OutTable->param1 = MyTable->param1;
            OutTable->ulMask = MyTable->ulMask;
            CLen = 0;
            Cx1 = (char *) MyTable->delim;
            Cx2 = (char *) OutTable->delim;
            while ((*Cx2++ = LOWERCASE(*Cx1++)) != '\0') CLen++;
            OutTable->size = CLen;
         }
         return (1);
      }
  
      /*----------------------------------------------------------------+
      !  case -3 .. Called once at shutdown. Can be used for cleanup.   ! 
      +----------------------------------------------------------------*/
      case -3:
      {  
         if (Common == NULL)  return 1;
          
               /*--------------------------------------------------+
               !  Free any GLOBAL MEMORY blocks we allocated.      !
               +--------------------------------------------------*/
          WatchClose(Common);
          Free(Common);
          *lpLong = 0;
          return (1);
      }
  }
  /*------------------------------+
  !   End of the Little Switch    !
  +------------------------------*/
  
     /***************************************************
     *                                                  *
     *  Before processing any other functions we        *
     *  must initialize the Common data area.           *
     *                                                  *
     *  (We use this area to pass information between   *
     *   called functions.)                             *
     *                                                  *
     ***************************************************/
  Common->Return          = ' ';
  Common->ErrorLevel      = -1; 
  Common->Array.lpBuffer  = NULL;          
  Common->DllServices     = lpViperStruct->lpfnDllServices;
  Common->hBatData        = lpViperStruct->hBatData;

     /***************************************************
     *                                                  *
     *    The "Big Switch"                              *
     *                                                  *
     *    Select an extender function.                  *
     *    The "CmdId" tells which one we want.          *
     *                                                  *
     *                                                  *
     ***************************************************/
  switch (CmdId)
  {
    case 10:
        WatchVersion  (Common, 
                       (nArgCount > 0) ? lpvvArg[0].x : 0
                      );
        break;
    
    case 11: 
        WatchOpen     (Common, 
                       lpvvArg[0].lpstr, // Directory
                       lpvvArg[1].x,     // Recursive
                       lpvvArg[2].lpstr, // Only
                       lpvvArg[3].lpstr  // Ignore
                      );
        break;
    
    case 12:
        WatchWait     (Common, 
                       lpvvArg[0].x, 
                       ((nArgCount > 1) ? lpvvArg[1].x : 0)
                      );
        break;
    
    case 13: 
        WatchClose    (Common);
        break;
 
    default: 
        SetError(Common, 1002, -1, "", "");
        break;
  }                   
  
  /*-------------------------------------------+
  !                                            !
  !         End of the Big Switch              ! 
  !                                            !
  +-------------------------------------------*/
           

 
  /***********************************************************
  *                                                          *
  *  Return the value, free allocated memory, and return.    *
  *                                                          *
  ***********************************************************/
  switch (Common->Return)         
  { 
      case 'A':   // Return an array
    {
       lpViperReturn->inittype = VARTYPE_ARRAY;
       lpViperReturn->flt      = Common->Array.flt;
       lpViperReturn->x        = Common->Array.x;
       lpViperReturn->lpstr    = Common->Array.lpBuffer;
       Ret = lpViperCallback(RESULT_SUCCESS, ViperInterfaceVersion, lpViperStruct, lpViperReturn);
       break;
    }
  
    case 'L':  // Return a long integer
    {
       lpViperReturn->inittype=VARTYPE_INT;
       lpViperReturn->x=Common->Long;
       Ret = lpViperCallback(RESULT_SUCCESS, ViperInterfaceVersion, lpViperStruct, lpViperReturn);
       break;
    }   
    
    case 'S':   // Return a string
    {
       lpViperReturn->inittype=VARTYPE_STRING;
       lpViperReturn->lpstr=Common->String;
       Ret = lpViperCallback(RESULT_SUCCESS, ViperInterfaceVersion, lpViperStruct, lpViperReturn);    
       break;
    }

    case 'F':   // Return a floating point number
    {
       lpViperReturn->inittype=VARTYPE_FLOATNUM;
       lpViperReturn->flt=Common->Float;
       Ret = lpViperCallback(RESULT_SUCCESS, ViperInterfaceVersion, lpViperStruct, lpViperReturn);
       break;
    }   
        
    case 'E':  // Return an error message
    {
       lpViperReturn->inittype=VARTYPE_STRING; 
       lpViperReturn->lpstr = Common->String;
       Ret = lpViperCallback(Common->ErrorLevel, ViperInterfaceVersion, lpViperStruct, lpViperReturn);
       break;
    }
    
    default:  // Invalid return type.  (This better not happen!)
    {
       LoadString(hinstDLL, 900, Common->Buffer, 128);
       lpViperReturn->inittype=VARTYPE_STRING; 
       lpViperReturn->lpstr = Common->Buffer;
       Ret = lpViperCallback(-3, ViperInterfaceVersion, lpViperStruct, lpViperReturn);
       break;
    }  
  }

  return Ret;
}
  /************************************
   *                                  *
   *   End of program WILExtender2    *
   *                                  *
   ***********************************/



/***********************************************************************************
************************************************************************************
**                                                                                **
**                                                                                **
**                                                                                **
**                                                                                **
**                                                                                **
**                                                                                **
**                                                                                **
**        The routines called to perform the extender functions.                  **
**                                                                                **
**                                                                                **
**                                                                                **
**                                                                                **
**                                                                                **
**                                                                                **
**                                                                                **
************************************************************************************
***********************************************************************************/



/************************************************************************
*  Extender Function...  WatchVersion                                   *
*                                                                       *
*             Return extender version information.                      *
*                                                                       *
*              Parameter 0 = Version                                    *
*                        1 = Total Functions                            *
*                        2 = Total Constants                            *
************************************************************************/
void  WatchVersion(COMMON *Common, long Request)
{
    long  N;
    long  Count;
    char *Bx;
    UINT  uType;

    if (Request == 0)             // Request 0 = Extender Version
    {
         char Work[24];
         LoadString(hinstDLL, 1001, Work, 24);
 
         N = 0;
         Bx = Work;
         while (*Bx != '\0')
         {
             N = (N * 10) + (*Bx - '0');
	         Bx++;
         }
         Common->Return = 'L';
         Common->Long = N;
         return;
    }

    if (Request == 1)                      // Request 1 = total functions
    {
        uType = FUNCTION;
    }
    else
    {
        if (Request == 2)                  // Request 2 = total constants
        {
            uType = CONSTANT;
        }
	else
        {                                  // Invalid parameter
            SetError(Common, 1003, -1, "", ""); 
            return;
        }
    }

    Count = 0;
    for (N = 0; CommandExtTable[N].ident != -1; N++)
    {
       if ((UINT) CommandExtTable[N].type == uType) Count++;
    }
    Common->Return = 'L';
    Common->Long = Count;
    return;
}

/***************************************************************************
*                                                                          *
*     WatchOpen(Common, Directory, Recursive, DoOnly, DontDo)              *
*                                                                          *
*      Directory.... The directory to be watched.                          *
*      Recursive.... TRUE = Do subdirectories, too.                        *
*      DoOnly....... Tab Delimited list. Watch only these files.           *
*      DontDo....... Tab Delimited list. Ignore these files.               *
*                                                                          *
*     Files in the DoOnly and DontDo list may contain wildcards.           *
*                                                                          *
***************************************************************************/
void  WatchOpen(COMMON *Common, 
                char   *Directory, 
                BOOL    Recursive, 
                char   *OnlyList, 
                char   *IgnoreList)
{
    HANDLE Handle;
    WATCH  *Watch;
    char   *Dx;
    long   Len;

    /*---------------------------------------------------------+
    !   We only accept a limited number of directories.        !
    !   (The limit can be increased - change the #define       !
    !   WATCHMAX value at the beginning of the program.)       !
    +---------------------------------------------------------*/
    if (Common->WatchCount == WATCHMAX)
    {
        char Number[24];
        wsprintf(Number,"%d",WATCHMAX);
        SetError(Common, 1004, -1, Number, "");
        return;
    }
    
    /*---------------------------------------------------------+
    !   Make sure we have a fully qualified file name.         !
    +---------------------------------------------------------*/
    Len = GetFullPathName(Directory, sizeof(Common->Path), Common->Path, &Dx);
    if (Len == 0)
    {                                      
        long Err;
        char Number[24];
        Err = GetLastError();
        wsprintf(Number,"%d",Err);
        SetError(Common, 1010, -1, Number, Directory);
        return;
    }   

    /*---------------------------------------------------------+
    !  Reject path names that are more than MAX_PATH long.     !
    +---------------------------------------------------------*/
    if (Len > MAX_PATH)
    {                                      
        char Number[24];
        wsprintf(Number,"%d",MAX_PATH);
        SetError(Common, 1007, -1, Number, "");
        return;
    }
    
    /*---------------------------------------------------------+
    !  See if the directory exists.                            !
    +---------------------------------------------------------*/    
    Dx = Common->Path + Len;                // Append "\*"
    if (*(Dx-1) != '\\') *Dx++ = '\\';      // to the directory;
    *Dx++ = '*';
    *Dx = '\0';
                                            // Scan the directory;
    Handle = FindFirstFile(Common->Path, &Common->w32find);    
    if (Handle == INVALID_HANDLE_VALUE)
    {    
        long E;
        switch (GetLastError())
        {
           case 5:
              E = 1009;   // Error 5 = Permission denied
              break;
           default:
              E = 1005;   // otherwise, assume directory doesn't exist.
              break;
        }                
        SetError(Common, E, -1, Directory, "");
        return;
    }
    FindClose(Handle);


    /*---------------------------------------------------------+
    !   Get a WATCH block big enough for this directory name.  !
    +---------------------------------------------------------*/   
    Watch = Allocate(Common,sizeof(WATCH) + Len + 2);
    if (Common->Return == 'E') return;
       
    /*---------------------------------------------------------+
    !   Copy the directory name into the Watch Block.          !
    !   We require a trailing slash at the end of the path.    !
    +---------------------------------------------------------*/
    lstrcpy(Watch->Dir,Common->Path);
    Watch->Dir[Len++] = '\\';
    Watch->Dir[Len] = '\0';    

    /*---------------------------------------------------------+
    !  Initialize the remainder of the WATCH block.            !
    +---------------------------------------------------------*/
    Watch->DirLen      = Len;  
    Watch->Recursive   = Recursive;
    Watch->Only        = NULL;
    Watch->Ignore      = NULL;
    Watch->Files       = NULL;
    Watch->ChangeFlag  = FALSE; 

    /*---------------------------------------------------------+
    !  Build the ONLY and IGNORE lists.                        !
    +---------------------------------------------------------*/       
    Watch->Only = BuildOnlyIgnore(Common, OnlyList);
    if (Common->Return == 'E')
    {
        FreeWatch(Watch);
        return;
    }
    
    Watch->Ignore = BuildOnlyIgnore(Common, IgnoreList);
    if (Common->Return == 'E')
    {
        FreeWatch(Watch);
        return;
    }
    
    /*-------------------------------------------------------+
    !   Get a list of files under this directory.            !
    +-------------------------------------------------------*/
    Watch->Files = CreateFileChain(Common, Watch);
    if (Common->Return == 'E')
    {
        FreeWatch(Watch);
        return;
    }
 
    /*-------------------------------------------------------+
    !   Ask the system to monitor this directory.            !
    !   (Note... we must string the trailing '\' from        !
    !   the file name before we make the API call.)          !
    +-------------------------------------------------------*/ 
    lstrcpy(Common->Path,Watch->Dir);
    Common->Path[Len-1] = '\0';  
    Handle = FindFirstChangeNotification(Common->Path,TRUE,
                                         FILE_NOTIFY_CHANGE_FILE_NAME
                                       | FILE_NOTIFY_CHANGE_DIR_NAME
                                       | FILE_NOTIFY_CHANGE_LAST_WRITE);
        
    if (Handle == INVALID_HANDLE_VALUE)
    {
       char Number[24];
       long Err;
       
       Err = GetLastError();
       if (Err == 50)            // 50 = Network request not supported.
       {
           SetError(Common,1011,-1,Directory,"");
           return;
       }
       wsprintf(Number,"%d",Err);
       FreeWatch(Watch);
       SetError(Common,1008,-1,Number,Directory);
       return;
    }
        
    /*-------------------------------------------------------+
    !   We're done.  Save the watch item and exit.           !
    +-------------------------------------------------------*/
    Common->Watches[Common->WatchCount] = Watch;
    Common->Handles[Common->WatchCount] = Handle;
    Common->WatchCount++;
     
    Common->Return = 'L';
    Common->Long = 1;
    return;
}
  
/***************************************************************************
*                                                                          *
*     WatchWait(Common, Timeout)                                           *
*                                                                          *
*      Timeout in MS  (1000 = 1 second.  -1 = No timeout.)                 *
*                                                                          *
*      DetailFlag,  TRUE = return an array with file name and info.        *
*                   FALSE = return only the file name.                     *
*                                                                          *
*      Return when a file has been created or modified,  or when the       *
*      timeout interval has expired.                                       *
*                                                                          *
***************************************************************************/
void  WatchWait(COMMON *Common, long Timeout, BOOL DetailFlag)
{
     long       EndTicks;
     long       HandleCount;
     long       WaitStatus;
     long       Interval;
     BOOL       X;
     HANDLE    *HandleTable;
     MSG        Msg;
 
     /*--------------------------------------------------------+
     !   Can't call this unless we've called WatchOpen         !
     !   to define at least one watch handle.                  !
     +--------------------------------------------------------*/   
     HandleCount = Common->WatchCount;
     HandleTable = &Common->Handles[0];
     if (HandleCount == 0)
     {
         SetError(Common, 1006, -1, "","");
         return;
     }
     /*--------------------------------------------------------+
     !  We want to check every time we're called to see if     !
     !  there are more changes.  If we notice that there are   !
     !  unreported changes waiting in the queue, set the       !
     !  timeout to zero so we will do our check and  (if       !
     !  there aren't any changes) return immediately.          !
     +--------------------------------------------------------*/
     if (Common->ChangeQueue != NULL)  Timeout = 0;  
     
     /*--------------------------------------------------------+
     !   Test timeout value.                                   !
     !          0 = Return immediately.                        !
     !         -1 = never timeout.                             !
     !  any other = compute timeout interval.                  !
     +--------------------------------------------------------*/     
     if (Timeout == 0 || Timeout == -1)
     {
         Interval = Timeout;
         Timeout = 0;
     }
     else
     {
        EndTicks = (long) GetTickCount() + Timeout;  // End time in ticks    
     }

     /*----------------------------------------------------------+
     !   Test for files until the timeout occurs, or             !
     !   some file changed are reported.                         !
     +----------------------------------------------------------*/
     while(1)
     {
          /*-----------------------------------------------------------+
          !  Compute the timeout interval.                             !
          !      If timeout was 0 or -1, we've already set the         !
          !      interval, and this is not required.                   !
          +-----------------------------------------------------------*/
          if (Timeout)
          {
             Interval = EndTicks - (long) GetTickCount();
             if (Interval < 1) Interval = 0;
          }
          
          /*----------------------------------------------------------+
          !   Wait... call will exit when (1) Something changed.      !
          !                               (2) Timeout happenes.       !
          !                               (3) System Message arrives. !
          +----------------------------------------------------------*/
          WaitStatus = MsgWaitForMultipleObjects(HandleCount,HandleTable,0,Interval,255);
          
          /*---------------------------------------------------------+
          !  If we timed out, return a TIMOUT status to the user.    !
          +---------------------------------------------------------*/
          if (WaitStatus == WAIT_TIMEOUT) break;

          /*---------------------------------------------------------+
          !  If we got a change notification, Check repeatedly       !
          !  (with timeout = zero) to see if there are any more.     !
          !  Then scan the directories to see what the changes are.  !
          +---------------------------------------------------------*/
          if (WaitStatus < HandleCount)
          {
              while (TRUE)
              {
                 Common->Watches[WaitStatus]->ChangeFlag = TRUE;
                 FindNextChangeNotification(Common->Handles[WaitStatus]);  
                 WaitStatus = MsgWaitForMultipleObjects(HandleCount,HandleTable,0,0,255);
                 if (WaitStatus == WAIT_TIMEOUT) break;
                 if (WaitStatus < HandleCount) continue;
                 while (1)
                 {
                   if (PeekMessage(&Msg,NULL,0,0,PM_REMOVE) == 0) break;
                   DispatchMessage(&Msg);
                 }                
              }
              /*-------------------------------------------------------+
              !  When copying a file into a new folder, we often get   !
              !  both a NEW FILE and a CHANGE notice. (NEW when the    !
              !  file is created, CHANGE when the copy finishes and    !
              !  the modify time gets updated.)  To cut down on this   !
              !  we sleep for 1/10 of a second after a notice arrives  !
              !  before we scan the directory. This allows most small  !
              !  files time to finish updating before we see them.     !
              +-------------------------------------------------------*/
              Sleep(100);
              
              /*-------------------------------------------------------+
              !   Scan the directory looking for changed files.        !
              +-------------------------------------------------------*/
              for (X=0; X<HandleCount; X++)
              {   
                 if (Common->Watches[X]->ChangeFlag)
                 {                
                   LookForChanges(Common, Common->Watches[X]);
                   Common->Watches[X]->ChangeFlag = FALSE;
                   if (Common->Return == 'E') return;
                 }
              }
              
              /*-------------------------------------------------------+
              !   If we found any changes, return.  (There won't       !
              !   always be changes. We might have received a notice   !
              !   because a file was deleted.)                         !
              +-------------------------------------------------------*/
              if (Common->ChangeQueue != NULL) break;
          }
          
          /*---------------------------------------------------------+
          !  Process system messages (if any)                        !
          +---------------------------------------------------------*/
          while (1)
          {
              if (PeekMessage(&Msg,NULL,0,0,PM_REMOVE) == 0) break;
              DispatchMessage(&Msg);
          }
     }
              
     if (Common->ChangeQueue == NULL)  ReturnTimeoutToUser(Common, DetailFlag);
     else                              ReturnChangesToUser(Common, DetailFlag);
     return;
}


/***************************************************************************
*                                                                          *
*     WatchClose(Common)                                                   *
*                                                                          *
*      Terminates the watch and frees all allocated structures.            *
*                                                                          *
***************************************************************************/
void  WatchClose(COMMON *Common)
{
    long X;
    
    /*-------------------------------------------------------------+
    !  Tell windows to stop watching the directories,              !
    !  then free the directory structures.                         !
    +-------------------------------------------------------------*/
    for (X=0; X<Common->WatchCount; X++)
    {
        if (Common->Handles[X] != INVALID_HANDLE_VALUE)
        {
           FindCloseChangeNotification(Common->Handles[X]);
           Common->Handles[X] = INVALID_HANDLE_VALUE;
        }
        FreeWatch(Common->Watches[X]);
        Common->Watches[X] = NULL;
    }
    Common->WatchCount = 0;
        
    /*------------------------------------------------------------+
    !  Free any leftover CHANGED structures.                      !
    +------------------------------------------------------------*/
    FreeLinkedList((LINKED *) Common->ChangeQueue);
    Common->ChangeQueue = NULL;
    
    Common->Return = 'L';
    Common->Long = 1;
}





/***************************************************************************
*                                                                          *
*   Build a chain of ONLY or IGNORE items.                                 *
*                                                                          *
*   Input is a delimited list of names.  It may be delimited by            *
*   tabs or vertical-bars ("|").                                           *
*                                                                          *
***************************************************************************/
LINKED *BuildOnlyIgnore(COMMON *Common, char *List)                  
{
    char   *End;
    char   *Start;
    long   Len;
    LINKED *Root;
    LINKED *New;
        
    Root = NULL;
    while (1)
    {
        /*---------------------------------------------------------+
        !  Skip spaces and delimiters to find the next name.       !
        +---------------------------------------------------------*/
        while (*List == ' ' || *List == '\t' || *List == '|') List++;
        if (*List == '\0')  break;     // End of list.
        Start = List;
                
        /*---------------------------------------------------------+
        !  Look for a delimiter at the end of the name.            !
        +---------------------------------------------------------*/                
        while (!(*List == '\t' || *List == '|' || *List == '\0')) List++;
                      
        /*---------------------------------------------------------+
        !  Skip trailing spaces.                                   !
        +---------------------------------------------------------*/  
        End = List;     
        while (*(End-1) == ' ') End--;

        /*---------------------------------------------------------+
        !  Create a block to hold this name.                       !
        +---------------------------------------------------------*/  
        Len = End - Start;
        New = Allocate(Common,sizeof(LINKED) + Len);
        if (Common->Return == 'E')
        {
            FreeLinkedList(Root);  // If error, free what we've done
            Root = NULL;           // and return a null adddress.
            break;
        }
        lstrcpyn(New->Buffer,Start,Len+1);
        New->Next = Root;
        Root = New;
    }
    return Root;  // Return address of first item in the linked list.
}

/***************************************************************************
*                                                                          *
*              Allocate Memory.                                            *
*                                                                          *
***************************************************************************/
void *Allocate(COMMON *Common, long Size)
{
     void *Memory;     
     Memory = (void *) GlobalAlloc(GMEM_FIXED,Size);

     if (Memory == NULL)
     {                                   // Can't get the memory?
         SetError(Common,1999,-3,"",""); // This is a fatal error.
         return NULL;
     }
    
#if DEBUGALLOC == 1  
     GlobalAllocCount++;
#endif
          
     return Memory;
}

/***************************************************************************
*                                                                          *
*              Free memory.                                                *
*                                                                          *
***************************************************************************/
void Free(HGLOBAL Memory)
{
    if (Memory == NULL) return;
    Memory = GlobalFree(Memory);
    
#if DEBUGALLOC == 1 
    if (Memory == NULL)  GlobalAllocCount--;
#endif
    
    return; 
}

/**************************************************************************
*                                                                         *
*   Free all the memory for a WATCH block.                                *
*                                                                         *
**************************************************************************/
void  FreeWatch (WATCH *Watch)
{
     FreeLinkedList((LINKED *) Watch->Only);
     FreeLinkedList((LINKED *) Watch->Ignore);
     FreeLinkedList((LINKED *) Watch->Files);
     Free(Watch);
     return;
}

/**************************************************************************
*                                                                         *
*   Free all the memory blocks in an "only chain".                        *
*                                                                         *
**************************************************************************/
void  FreeLinkedList(LINKED *Linked)
{
    LINKED *Hold;

    while (Linked != NULL)
    {
        Hold = Linked->Next;
        Free(Linked);
        Linked = Hold;
    }
    return;
}


   
/**********************************************************************
*                                                                     *
*    Case insensitive String Compare.                                 *
*                                                                     *
*    Returns   -1 ... First < Second                                  *
*               0 ... First = Second                                  *
*               1 ... First > Second                                  *
*                                                                     *
*                                                                     *
**********************************************************************/
long CaseComp(char *First, char *Second)
{
   char A;
   char B;
   while (1)
   {
        A = LOWERCASE(*First++);
        B = LOWERCASE(*Second++);
        if (A < B) return -1;
        if (A > B) return 1;
        if (A == '\0') return 0;
   }
}

/****************************************************************************
*                                                                           *
*  Compare two strings, one of which can contain wild cards.                *
*                                                                           *
*    Wild cards are ...  * = Any number of characters                       *
*                        ? = Any single character.                          *                        *
*                                                                           *
*  Returns TRUE if the names match.                                         *
*                                                                           *
****************************************************************************/  
BOOL WildComp(char *String, char *Wild)
{    
    char Cw;
    
    while (1)
    {    
         Cw = LOWERCASE(*Wild);                 // Character of wildcard.          
         if (Cw == '\0')                        // If end of wildcard
         {                                      // is also end of string,
             if (*String == '\0') return TRUE;  // then we match.
             return FALSE;
         }
         
         if (Cw == '*')            // Is it *?
         {
             ++Wild;                          // Bump wildcard. A * at
             if (*Wild == '\0') return TRUE;  // the end matches everything.
             
             while (TRUE)          // See if we can match the rest
             {                     // of the wildcard anywhere 
                  if (WildComp(String, Wild)) return TRUE;
                  if (*String == '\0')        return FALSE;
                  ++String;
             }
         }
         
         if (Cw == '?')             // A question mark matches
         {                          // any single character
             if (*String == '\0') return FALSE;
             String++;
             Wild++;
             continue;
         }
                  // Anything else must match.

         if (LOWERCASE(*String++) != Cw) return FALSE;
         Wild++;
         continue;
    }
}        
            
/***************************************************************************
*                                                                          *
*     Debug Functions                                                      *
*                                                                          *
*      MBSN ... Display a decimal number.                                  *
*      MBSX ... Display a hexidecimal number.                              *
*      MBSS ... Display a string.                                          *
*                                                                          *
***************************************************************************/ 
 void FAR PASCAL MBSN(char *Title, long Number)
 {
     char Debug[128]; 
     wsprintf(Debug,"%d",Number);
     MessageBox(GetFocus(), Debug, Title, MB_OK | MB_SETFOREGROUND);
     return;
 }

 void FAR PASCAL MBSX(char *Title, long Number)
 {
     char Debug[128]; 
     wsprintf(Debug,"x%08lX",Number);
     MessageBox(GetFocus(), Debug, Title, MB_OK | MB_SETFOREGROUND);
     return;
 }
 
void FAR PASCAL MBSS(char *Title, char *Text)
{
     MessageBox(GetFocus(), Text, Title, MB_OK | MB_SETFOREGROUND);
}


 
/***************************************************************************
****************************************************************************
**                                                                        **
**                                                                        **
**                                                                        **
**                                                                        **
**                                                                        **
**   Functions used to build a chain of file names in memory.             **
**                                                                        **
**                                                                        **
**   The chain contains all the file names under a given directory,       **
**   including subdirectories,  sorted in descending sequence.            **
**                                                                        **
**                                                                        **
**                                                                        **
**                                                                        **
**                                                                        **
****************************************************************************
***************************************************************************/



/************************************************************************
*                                                                       *
*  Creates a file chain for one of the WATCH directories.               *
*                                                                       * 
*  Returns a pointer to the first FILECHAIN block in the chain,         *
*  or NULL if there was an error.                                       *
*                                                                       *
*  The only error that could occur here is "unable to allocate memory". *
*                                                                       *
************************************************************************/
FILECHAIN *CreateFileChain(COMMON *Common, WATCH *WatchMe)
{
   FILECHAINWORK  Work;


   Work.Common    = Common;
   Work.Path      = Common->Path;
   Work.FileInfo  = &Common->w32find;
   Work.Recursive = WatchMe->Recursive;
   Work.Only      = WatchMe->Only;
   Work.Ignore    = WatchMe->Ignore;
   Work.Root.Next = NULL;
   
   lstrcpy(Work.Path,WatchMe->Dir);
   Work.PathAboveBase = Work.Path + WatchMe->DirLen;
      
   GetFilesRecursive(&Work, Work.PathAboveBase);  // Get files
   if (Common->Return == 'E')
   {                                              // If there was an error
      FreeLinkedList((LINKED*)Work.Root.Next);    // free everything
      return NULL;                                // and return NULL.
   }
   return Work.Root.Next;                // No error. Return chain address.
}


/***************************************************************************
*                                                                          *
*  Called by the "CreateFileChain" routine.                                *
*                                                                          *
*  Walk the current directory.                                             *
*  When we find a file, add its name to a  chain of filenames.             *
*  When we find a directory, call ourself recursively and keep looking.    *
*                                                                          *
*  The MASK argument is a filename which may (and usually will) contain    *
*  wildcards.  Only file names that match the mask will be saved.          *
*                                                                          *
*  Ignore is a chain of wildcard masks telling us what files to ignore.    *
*                                                                          *
*  Normally returns TRUE.                                                  *
*  Returns FALSE if unable to get memory to build the chain.               *
*                                                                          *
***************************************************************************/
void GetFilesRecursive(FILECHAINWORK *Work, char *Append)
{                        
    HANDLE           hFIND;         // Handle for FindNextFile 
    BOOL             Keep;
    char            *Fx;
    char            *Ax; 
    char            *FileName;
    WIN32_FIND_DATA *Info;
    
    Info         = Work->FileInfo;
    FileName     = Info->cFileName;
    if (*(Append-1) != '\\') *Append++ = '\\';
    
    /*-------------------------------------------------------+
    !   Search using a mask of "*" to return everything.     !
    +-------------------------------------------------------*/
    *Append = '*'; 
    *(Append+1) = '\0';
    
    hFIND = FindFirstFile(Work->Path, Info); 
    if (hFIND == INVALID_HANDLE_VALUE)  return;
  
    /*-------------------------------------------------------+
    !   Loop to process the file names.                      !
    +-------------------------------------------------------*/
    while (1)
    {
        do  // Put code inside a "do" so we can use "break" to skip to the end.
        {
            //-------------------------------------------------------------------
            //  "." and ".." are dummy directory names returned by the
            //  find file routines.  "." means the current directory
            //  and '.." means the parent directory.  We ignore both.
            //  Make sure OK=TRUE to show that there was not an error.
            //-------------------------------------------------------------------
            if (FileName[0] == '.')
            {
                if (FileName[1] == '\0') break;
                if (FileName[1] == '.' && FileName[2] == '\0') break;
            }

            //-------------------------------------------------------------------
            //  For a directory (if Recursive is set) call ourself again
            //  to process the subdirectory.
            //-------------------------------------------------------------------
            if (Info->dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
            {
                if (Work->Recursive)                     // Do Subdirectories?
                {
                   Fx = FileName;                        // Append the name              
                   Ax = Append;                          // to the current path
                   while (*Fx != '\0') *Ax++ = *Fx++;
                   *Ax++ = '\\';
                   GetFilesRecursive(Work, Ax);
                }
                break;
            }
            
            /*--------------------------------------------------------+
            !  Check the ONLY and IGNORE lists.                       !
            +--------------------------------------------------------*/	
            if (Work->Only == NULL)
                Keep = TRUE;
            else
                Keep = FoundInLinkedList(FileName, Work->Only);
            
            if (Keep)
              if (Work->Ignore != NULL)
                Keep = !FoundInLinkedList(FileName, Work->Ignore);
         
            if (!Keep) break;
            
            /*--------------------------------------------------------+
            !  If the file passed the tests, keep it.                 !
            +--------------------------------------------------------*/	    
            Fx = FileName;                        // Append the name
            Ax = Append;                          // to the current path
            while (*Fx != '\0') *Ax++ = *Fx++;
            *Ax = '\0';
            Work->LengthAboveBase = Ax - Work->PathAboveBase;
            
            FileChainInsert(Work);                // Insert into chain.
            break;
     	}while (0);  // end of dummy DO loop
    	
    	if (Work->Common->Return == 'E') break;   // Exit if error happened
    	if (!FindNextFile(hFIND,Info)) break;     // else get the next file. 
    }
    
    FindClose(hFIND);
    return;     
}

/**************************************************************************
*    Compare a file name to the items in an ONLY or IGNORE list.          *
*                                                                         *
*    If the list is empty we select all files, so return TRUE.            *
*    Otherwise, return TRUE if the name matches an item in the list.      *
*                                                                         *
**************************************************************************/
BOOL FoundInLinkedList(char *Name, LINKED *List)
{
     if (List != NULL)
     {
        while (List != NULL)
        {
           if (WildComp(Name,List->Buffer)) return TRUE;
           List = List->Next;
        }
     }
     return FALSE;
}

      
/**************************************************************************
*                                                                         *
*  Called by the "GetFilesRecursive" routine                              *
*                                                                         *
*  Insert a new "fileNameBlock" into a chain of fileNameBlocks.           *
*  The block is inserted in alphanumeric order  (case insensitive.)       *
*                                                                         *
*  Note... Since files are usually returned in more-or-less sorted order  *
*  we can save many compares by sorting in descending order.              *
*                                                                         *
*  Return FALSE if we can't get memory to hold the name.                  *
*  Return TRUE  on a successful completion.                               *
*                                                                         *
**************************************************************************/
void FileChainInsert(FILECHAINWORK *Work)
{
    FILECHAIN *Previous;
    FILECHAIN *Current;
    FILECHAIN *New;
    long       Match;

    Previous = &Work->Root;                            // Search the chain
    while (1)
    {
       if ((Current = Previous->Next) == NULL) break;       // Attach at end of chain
       Match = CaseComp(Current->Name,Work->PathAboveBase); // Test sort sequence.      
       if (Match == -1) break;                              // Attach at sorted location.
       Previous = Current;
    }
    
    New = (FILECHAIN *) Allocate(Work->Common, sizeof(FILECHAIN) + Work->LengthAboveBase + 1);
    if (Work->Common->Return == 'E') return;
       
    lstrcpy(New->Name,Work->PathAboveBase);         // save the name.
    New->Size     =  Work->FileInfo->nFileSizeHigh;
    New->Size     =  New->Size << 32;
    New->Size     += Work->FileInfo->nFileSizeLow;
    New->Modified =  Work->FileInfo->ftLastWriteTime.dwHighDateTime;
    New->Modified =  New->Modified << 32;
    New->Modified += Work->FileInfo->ftLastWriteTime.dwLowDateTime; 
    New->Created  =  Work->FileInfo->ftCreationTime.dwHighDateTime;
    New->Created  =  New->Created << 32;
    New->Created  += Work->FileInfo->ftCreationTime.dwLowDateTime; 
    
    New->Next = Previous->Next;                // Insert the new file
    Previous->Next = New;                      // between PREVIOUS and
    return;                                    // PREVIOUS->NEXT
    
} 



/***************************************************************************
*                                                                          *
*                                                                          *
*      Compare an old list of files with a new one.                        *
*                                                                          *
*      Files that have been added or modified are saved in                 *
*      Common->ChangeQueue                                                 *
*                                                                          *
*                                                                          *
***************************************************************************/
void  LookForChanges(COMMON *Common, WATCH *Watch)
{
    BOOL       Error;
    
    FILECHAIN  NewRoot;
    FILECHAIN  OldRoot;
    FILECHAIN  KeepRoot;   
     
    FILECHAIN *Old;
    FILECHAIN *New;
    FILECHAIN *Keep;

    
    Error = FALSE;
    NewRoot.Next = NULL;
   

    OldRoot.Next  = Watch->Files;
    NewRoot.Next  = CreateFileChain(Common, Watch);
    if (Common->Return == 'E') return;
   
    Keep = &KeepRoot;
    KeepRoot.Next = NULL; 
        
    while (1)
    {
         Old = OldRoot.Next;
         New = NewRoot.Next;
         if (Old == NULL) break;   // At end of either list?
         if (New == NULL) break;
         
         switch (CaseComp(Old->Name,New->Name))
         {
             /*-------------------------------------+
             !  File names match                    !
             +-------------------------------------*/
             case 0:
             {
                 NewRoot.Next = New->Next;         // Remove both blocks
                 OldRoot.Next = Old->Next;         // from their chains.
              
                 if (Old->Modified != New->Modified)  // If the filetime chanbged
                 {                                    // save the file infor
                    SaveChanged(Common,Watch,New,0);  // for return to the user.
                 }

                 New->Next = NULL;                    // Keep the new file  
                 Keep->Next = New;            
                 Keep = New; 
                   
                 Free(Old);                           // Throw out the old file
                 break; 
             }
         
             /*-----------------------------------------+
             !  New file not on OLD list.               !
             !  (File was created. Keep it.)            !
             +-----------------------------------------*/
             case -1:           
             {
                 NewRoot.Next = New->Next;           // Remove from new chain
                 SaveChanged(Common, Watch, New, 1); // Save new-file info
                 
                 New->Next = NULL;
                 Keep->Next = New;                  // Add to KEEP chain
                 Keep = New;
                 break;
             }
             /*-----------------------------------------+
             !  Old file not on NEW list.               !
             !  (File has been deleted. Drop it.)       !
             +-----------------------------------------*/                                                 // Old file not on new list.
             default:
             {                                  
                OldRoot.Next = Old->Next;         // Drop the OLD file
                Free(Old);
                break;
             } 
         }
    }
    
    /*------------------------------------------------+
    !  Were there any files left on the OLD list?     !
    !  They have been deleted... Drop them.           !
    +------------------------------------------------*/
    while (OldRoot.Next != NULL)
    {             
        Old = OldRoot.Next;
        OldRoot.Next = Old->Next;
        Free(Old);
    }
    
    /*------------------------------------------------+
    !  Were there any files left on the NEW list?     !
    !  They have been created.  Keep them.            !
    +------------------------------------------------*/    
    while (NewRoot.Next != NULL)
    {
         New = NewRoot.Next;                  // Remove from list
         NewRoot.Next = New->Next;
 
         SaveChanged(Common, Watch, New, 1);  // Save new file info. 
                   
         New->Next = NULL;
         Keep->Next = New;                    // add to keep list.       
         Keep = New;
     }
     
     Watch->Files = KeepRoot.Next;
     return;
}

/********************************************************************************
*                                                                               *
*   When we find a NEW or CHANGED file,  save the file information in the       *
*   Common->ChangeQueue.  Items in this queue will be returned to the caller.   *
*                                                                               *
********************************************************************************/
void SaveChanged(COMMON *Common, WATCH *Watch, FILECHAIN *File, BOOL NewFileFlag)
{
   long       LenFile;
   CHANGED    *Chg;

   if (Common->Return == 'E') return;

   /*-------------------------------------------------------+
   !   Get memory and create a new CHANGED block.           !
   +-------------------------------------------------------*/    
   LenFile = lstrlen(File->Name);   
   Chg = Allocate(Common, LenFile+sizeof(CHANGED));
   if (Common->Return == 'E') return;

   Chg->Watch       = Watch;
   Chg->Size        = File->Size;
   Chg->Created     = File->Created;
   Chg->Modified    = File->Modified;
   Chg->NewFileFlag = NewFileFlag;
   lstrcpy(Chg->Name,File->Name);   

   /*-------------------------------------------------------+
   !   Add the new item at the end of the change queue.     !
   +-------------------------------------------------------*/
   if (Common->ChangeQueue == NULL)
   {
      Chg->Next = Common->ChangeQueue;
      Common->ChangeQueue = Chg;
   }
   else
   {
      CHANGED *Last= Common->ChangeQueue;
      Chg->Next = NULL;
      while (Last->Next != NULL) Last = Last->Next;
      Last->Next = Chg;
   }
   return;
}
  

/********************************************************************************
*                                                                               *
*   When we find a NEW or CHANGED file,  save the file information in the       *
*   Common->ChangeQueue.  Items in this list will be returned to the caller.    *
*                                                                               *
********************************************************************************/
void ReturnChangesToUser(COMMON *Common, BOOL Details)
{
   ULONG64    L64;
   FILETIME   ft;
   SYSTEMTIME st;
   CHANGED   *Chg;
   LPARRVAR   Arr; 
   char      *Wx;
   char       Work[24];

   
   Chg = Common->ChangeQueue; 
  
   /*-----------------------------------------------------------+
   !  Concatenate base directory & file to make                 !
   !  a fully-qualified path.                                   !
   +-----------------------------------------------------------*/
   lstrcpy(Common->Path,Chg->Watch->Dir);
   lstrcpyn(Common->Path + Chg->Watch->DirLen, Chg->Name, 
             sizeof(Common->Path) - Chg->Watch->DirLen - 1);
   
   /*-----------------------------------------------------------+
   !  If user requested NO-DETAILS, return only the path.       !
   +-----------------------------------------------------------*/
   if (!Details)                       // If no details,
   {                                   // return just the path.
       Common->Return = 'S';
       Common->String = Common->Path;
       Common->ChangeQueue = Chg->Next;  
       Free(Chg);
       return;
   }
   
   /*------------------------------------------------------------+
   ! If user requested DETAILS, return the details in an array.  !
   +------------------------------------------------------------*/
   Common->Array.cDims   = 1;
   Common->Array.wDim[0] = 5;
   Common->Array.wDim[1] = 0;
   Common->Array.wDim[2] = 0;
   Common->Array.wDim[3] = 0;
   Common->Array.wDim[4] = 0;
   Common->Array.cFlags  = 0;
   Common->Array.lpBuffer = Common->Buffer;

   
   Arr = (LPARRVAR) Common->Array.lpBuffer;        // Array[0] = Path
   if (!ArrayStringPUT(Common, Arr, Common->Path)) return; 

   Arr += 1;                                       // Array[1] = NewFileFlag
   Arr->Ainittype = VARTYPE_INT;
   Arr->Ax = (long) Chg->NewFileFlag;
   
   Arr += 1;                                       // Array[2] = Size
   if (Chg->Size > 0x7FFFFFFF)
   {
       Work[23] = '\0';                            // If it's too big to 
       Wx = Work+22;                               // fit in a LONG variable
       L64 = Chg->Size;                           // convert it to a string.
       while (1)
       {
          *Wx = (L64 % 10) + '0';
          L64 = L64 / 10;
          if (L64 == 0) break;
          Wx--;
       }
       if (!ArrayStringPUT(Common, Arr, Wx)) return; 
    }
    else
    {  
       Arr->Ainittype = VARTYPE_INT;
       Arr->Ax = (long) Chg->Size;  
    } 

    Arr += 1;                                       // Array[3] = Creation date
   
    FileTimeToLocalFileTime((FILETIME *) &Chg->Created, &ft);
    FileTimeToSystemTime(&ft,&st);
    Work[0]  = ((st.wYear / 1000) % 10) + '0';
    Work[1]  = ((st.wYear / 100)  % 10) + '0';
    Work[2]  = ((st.wYear / 10)   % 10) + '0';
    Work[3]  = ( st.wYear         % 10) + '0';
    Work[4]  = ':';
    Work[5]  = ((st.wMonth / 10)  % 10) + '0';
    Work[6]  = ( st.wMonth        % 10) + '0';
    Work[7]  = ':';
    Work[8]  = ((st.wDay   / 10)  % 10) + '0';
    Work[9]  = ( st.wDay          % 10) + '0';
    Work[10] = ':';
    Work[11] = ((st.wHour  / 10)  % 10) + '0';
    Work[12] = ( st.wHour         % 10) + '0';
    Work[13] = ':';
    Work[14] = ((st.wMinute/ 10)  % 10) + '0';
    Work[15] = ( st.wMinute       % 10) + '0';
    Work[16] = ':';
    Work[17] = ((st.wSecond/ 10)  % 10) + '0';
    Work[18] = ( st.wSecond       % 10) + '0';   
    Work[19] - '\0';
    if (!ArrayStringPUT(Common,Arr,Work)) return;    
   
    Arr += 1;                                       // Array[4] = Modify Date
   
    FileTimeToLocalFileTime((FILETIME *) &Chg->Modified, &ft);
    FileTimeToSystemTime(&ft,&st);
    Work[0]  = ((st.wYear / 1000) % 10) + '0';
    Work[1]  = ((st.wYear / 100)  % 10) + '0';
    Work[2]  = ((st.wYear / 10)   % 10) + '0';
    Work[3]  = ( st.wYear         % 10) + '0';
    Work[4]  = ':';
    Work[5]  = ((st.wMonth / 10)  % 10) + '0';
    Work[6]  = ( st.wMonth        % 10) + '0';
    Work[7]  = ':';
    Work[8]  = ((st.wDay   / 10)  % 10) + '0';
    Work[9]  = ( st.wDay          % 10) + '0';
    Work[10] = ':';
    Work[11] = ((st.wHour  / 10)  % 10) + '0';
    Work[12] = ( st.wHour         % 10) + '0';
    Work[13] = ':';
    Work[14] = ((st.wMinute/ 10)  % 10) + '0';
    Work[15] = ( st.wMinute       % 10) + '0';
    Work[16] = ':';
    Work[17] = ((st.wSecond/ 10)  % 10) + '0';
    Work[18] = ( st.wSecond       % 10) + '0';   
    Work[19] - '\0';
    if (!ArrayStringPUT(Common,Arr,Work)) return;  
  
    Common->Return = 'A';
    Common->ChangeQueue = Chg->Next;  
    Free(Chg);
    return;
}


/********************************************************************************
*                                                                               *
*   Return null information to the user when the timeout expires                *
*                                                                               *
********************************************************************************/
void ReturnTimeoutToUser(COMMON *Common, BOOL Details)
{
   LPARRVAR Arr;
   
   Common->Path[0] = '\0';             // An empty string;
   
   if (!Details)                       // If no details,
   {                                   // return just the empty string.
       Common->Return = 'S';
       Common->String = Common->Path;
       return;
   }
                                       // If details,
   Common->Array.cDims   = 1;          // initialize an array
   Common->Array.wDim[0] = 5;
   Common->Array.wDim[1] = 0;
   Common->Array.wDim[2] = 0;
   Common->Array.wDim[3] = 0;
   Common->Array.wDim[4] = 0;
   Common->Array.cFlags  = 0;
   Common->Array.lpBuffer = Common->Buffer;

   Arr = (LPARRVAR) Common->Array.lpBuffer;        // Array[0] = The empty string
   if (!ArrayStringPUT(Common, Arr, Common->Path)) return; 

   Arr += 1;                                       // Array[1] 
   Arr->Ainittype = VARTYPE_INT;                   //          All other array
   Arr->Ax = 0;                                    //          items are zero.
   Arr += 1;                                       // Array[2] 
   Arr->Ainittype = VARTYPE_INT;
   Arr->Ax = 0;
   Arr += 1;                                       // Array[3] 
   Arr->Ainittype = VARTYPE_INT;
   Arr->Ax = 0;   
   Arr += 1;                                       // Array[4] 
   Arr->Ainittype = VARTYPE_INT;
   Arr->Ax = 0;   
   
   Common->Return = 'A';
   return;
}

/**************************************************************************
*    Put a string value into an array variable.                           *
*    Is passed the array pointer, and the string.                         *
*                                                                         *
*    Returns TRUE if successful.                                          *
*                                                                         *
*    If the operation fails, puts an error indicator in the Common block  *
*    and returns FALSE.                                                   *
**************************************************************************/
BOOL ArrayStringPUT(COMMON *Common, LPARRVAR lpArray, char *TheString) 
{
    lpArray->Ainittype = VARTYPE_STRING;  
    lpArray->Astrptr = (DWORD) (*Common->DllServices)(Common->hBatData, 42, TheString, "", 0, 0);
    if ( lpArray->Astrptr == 0)
    {
       SetError(Common, 1013, -3, "", "");
       return FALSE;
    } 
    return TRUE;
}


/****************************************************************************
*                                                                           *
*  When there is an error...                                                *
*                                                                           *
*   Read the error string from the resource file.                           *
*   Substitute in two parameters.                                           *
*   Set the flags to return the error message.                              *
*                                                                           *
****************************************************************************/
void SetError(COMMON *Common, long ErrNo, long ErrLevel, char *Param1, char *Param2)
{
    char Text[128];

    if (Common == NULL) return;
    
    LoadString(hinstDLL, ErrNo, Text, 128);
    wsprintf(Common->Buffer,Text,Param1,Param2);   
    Common->Return = 'E';
    Common->String = Common->Buffer;
    return;
}